#ifndef BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_
#define BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_

#include <string>

#include "fsfw/devicehandlers/CookieIF.h"
#include "fsfw/objectmanager/SystemObject.h"
#include "fsfw/osal/linux/BinarySemaphore.h"
#include "fsfw/returnvalues/HasReturnvaluesIF.h"
#include "fsfw/tasks/ExecutableObjectIF.h"
#include "fsfw_hal/linux/uart/UartComIF.h"
#include "linux/devices/devicedefinitions/PlocMPSoCDefinitions.h"
#ifdef XIPHOS_Q7S
#include "bsp_q7s/memory/SdCardManager.h"
#endif

/**
 * @brief   Helper class for supervisor of PLOC intended to accelerate large data transfers between
 *          the supervisor and the OBC.
 * @author  J. Meier
 */
class PlocSupvHelper : public SystemObject, public ExecutableObjectIF, public HasReturnvaluesIF {
 public:
  static const uint8_t SUBSYSTEM_ID = SUBSYSTEM_ID::PLOC_MPSOC_HELPER;

  //! [EXPORT] : [COMMENT] update failed
  static const Event SUPV_UPDATE_FAILED = MAKE_EVENT(0, severity::LOW);
  //! [EXPORT] : [COMMENT] update successful
  static const Event SUPV_UPDATE_SUCCESSFUL = MAKE_EVENT(1, severity::LOW);
  //! [EXPORT] : [COMMENT] Communication interface returned failure when trying to send the command
  //! ot the PLOC
  //! P1: Return value returned by the communication interface sendMessage function
  //! P2: Internal state of MPSoC helper
  static const Event SENDING_COMMAND_FAILED = MAKE_EVENT(2, severity::LOW);
  //! [EXPORT] : [COMMENT] Request receive message of communication interface failed
  //! P1: Return value returned by the communication interface requestReceiveMessage function
  //! P2: Internal state of MPSoC helper
  static const Event MPSOC_HELPER_REQUESTING_REPLY_FAILED = MAKE_EVENT(3, severity::LOW);
  //! [EXPORT] : [COMMENT] Reading receive message of communication interface failed
  //! P1: Return value returned by the communication interface readingReceivedMessage function
  //! P2: Internal state of MPSoC helper
  static const Event MPSOC_HELPER_READING_REPLY_FAILED = MAKE_EVENT(4, severity::LOW);
  //! [EXPORT] : [COMMENT] Did not receive acknowledgement report
  //! P1: Number of bytes missing
  //! P2: Internal state of MPSoC helper
  static const Event SUPV_MISSING_ACK = MAKE_EVENT(5, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor did not receive execution report
  //! P1: Number of bytes missing
  //! P2: Internal state of supervisor helper
  static const Event SUPV_MISSING_EXE = MAKE_EVENT(6, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor received acknowledgment failure report
  //! P1: Internal state of supervisor helper
  static const Event SUPV_ACK_FAILURE_REPORT = MAKE_EVENT(7, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor received execution failure report
  //! P1: Internal state of supervisor
  static const Event SUPV_EXE_FAILURE_REPORT = MAKE_EVENT(8, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor expected acknowledgment report but received space packet with other apid
  //! P1: Apid of received space packet
  //! P2: Internal state of supervisor helper
  static const Event SUPV_ACK_INVALID_APID = MAKE_EVENT(9, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor helper expected execution report but received space packet with other apid
  //! P1: Apid of received space packet
  //! P2: Internal state of supervisor helper
  static const Event SUPV_EXE_INVALID_APID = MAKE_EVENT(10, severity::LOW);

  PlocSupvHelper(object_id_t objectId);
  virtual ~PlocSupvHelper();

  ReturnValue_t initialize() override;
  ReturnValue_t performOperation(uint8_t operationCode = 0) override;

  ReturnValue_t setComIF(DeviceCommunicationIF* communicationInterface_);
  void setComCookie(CookieIF* comCookie_);

  /**
   * @brief   Starts update procedure
   *
   * @param file            File containing the update data
   * @param memoryId        ID of the memory where to write to
   * @param startAddress    Address where to write data
   *
   * @return  RETURN_OK if successful, otherwise error return value
   */
  ReturnValue_t startUpdate(std::string file, uint8_t memoryId, uint32_t startAddress);

  /**
   * @brief   Can be used to interrupt a running data transfer.
   */
  void stopProcess();

 private:
  static const uint8_t INTERFACE_ID = CLASS_ID::PLOC_SUPV_HELPER;

  //! [EXPORT] : [COMMENT] File accidentally close
  static const ReturnValue_t FILE_CLOSED_ACCIDENTALLY = MAKE_RETURN_CODE(0xA0);

  // Maximum number of times the communication interface retries polling data from the reply
  // buffer
  static const int RETRIES = 10000;

  struct Update {
    uint8_t memoryId;
    uint32_t startAddress;
    // Absolute name of file containing update data
    std::string file;
  };

  struct Update update;

  enum class InternalState { IDLE, UPDATE };

  InternalState internalState = InternalState::IDLE;

  BinarySemaphore semaphore;
#ifdef XIPHOS_Q7S
  SdCardManager* sdcMan = nullptr;
#endif
  uint8_t commandBuffer[supv::MAX_COMMAND_SIZE];

  bool terminate = false;

  /**
   * Communication interface responsible for data transactions between OBC and Supervisor.
   */
  UartComIF* uartComIF = nullptr;
  // Communication cookie. Must be set by the supervisor Handler
  CookieIF* comCookie = nullptr;

  ReturnValue_t performUpdate();
  ReturnValue_t handlePacketTransmission(SpacePacket& tc);
  ReturnValue_t sendCommand(SpacePacket& tc);
  ReturnValue_t receive(uint8_t* data, size_t* readBytes, size_t requestBytes);
  ReturnValue_t handleAck();
  ReturnValue_t handleExe();
  void handleAckApidFailure(uint16_t apid);
  void handleExeApidFailure(uint16_t apid);
  ReturnValue_t handleTmReception(SpacePacket* tmPacket, size_t remainingBytes);
  ReturnValue_t eraseMemory();
};

#endif /* BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_ */
